# <@LICENSE>
# Copyright 2023 IPQualityScore LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </@LICENSE>

=pod

=head1 NAME

IPQualityScore - email address reputation and validation

=head2 Description

This plugin integrates the IPQualityScore Email Validation API with SpamAssassin.
For each message checked by SpamAssassin, after the message has been parsed, the
plugin sends an API request to IPQualityScore with the sender's email address to
be checked against IPQualityScore's real-time email reputation and validation
system.

=head2 SYNOPSIS

Place the two files, IPQualityScore.pm and ipqualityscore.cf, into your
site-wide SpamAssassin directory, e.g. '/etc/mail/spamassassin'. Then add the
following to your init.pre:

    loadplugin IPQualityScore /path/to/IPQualityScore.pm

Edit the configuration file in ipqualityscore.cf to suit you. You must enter
your IPQualityScore API key on the line indicated in that file or the plugin
will not work.

Finally, update the SpamAssassin matching patterns and compile them. This
varies by platform. For example, if SpamAssassin is running as the system user
spamd and sa-update and sa-compile are located in /usr/bin/vendor_perl, then
you would run:

    sudo -u spamd /usr/bin/vendor_perl/sa-update &&
    sudo -u spamd /usr/bin/vendor_perl/sa-compile

=head2 CONFIGURATION

Configuration occurs in ipqualityscore.cf, for which a detailed template with
defaults that will be reasonable for most users has been provided. Please review
that file and make changes to it to suit you. A brief overview of the available
options is given here.

=over

=item ipqs_disable ( 0 | 1 ) (default: 0)

Setting ipqs_disable to 1 will disable the plugin.

=item fraud_score ( 0.0 - 100.0 ) (default: 75.0)

The fraud score threshold above which messages will trigger the
IPQS_FRAUD_SCORE rule, increasing that message's spam score.

=item fast ( 0 | 1 ) (default: 0)

When this parameter is enabled our API will not perform an SMTP check with the
mail service provider, which greatly increases the API speed. Syntax and DNS
checks are still performed on the email address as well as our disposable email
detection service. This option is intended for services that require decision
making in a time sensitive manner.

=item timeout ( 1 - 60 ) (default: 7)

Maximum number of seconds to wait for a reply from a mail service provider. If
your implementation requirements do not need an immediate response, we
recommend bumping this value to 20. Any results which experience a connection
timeout will return the "timed_out" variable as true.

=item strictness ( 0 | 1 | 2 ) (default: 0)

Sets how strictly spam traps and honeypots are detected by our system,
depending on how comfortable you are with identifying emails suspected of being
a spam trap. 0 is the lowest level which will only return spam traps with high
confidence. Strictness levels above 0 will return increasingly more strict
results, with level 2 providing the greatest detection rates.

=item abuse_strictness ( 0 | 1 | 2 ) (default: 0)

Set the strictness level for machine learning pattern recognition of abusive
email addresses with the "recent_abuse" data point. Default level of 0 provides
good coverage, however if you are filtering account applications and facing
advanced fraudsters then we recommend increasing this value to level 1 or 2.

=back

=head2 REQUIREMENTS

The following Perl libraries are required by this plugin:

=over

=item HTTP::Request

=item HTTP::Headers

=item JSON (e.g. JSON::XS)

=item LWP::UserAgent

=back

=head2 RULES

This plugin defines the following rules. Please reference ipqualityscore.cf for
details about each.

=over

=item IPQS_FRAUD_SCORE

    header   IPQS_FRAUD_SCORE eval:ipqs_fraud_score()
    score    IPQS_FRAUD_SCORE 1.0
    describe IPQS_FRAUD_SCORE IPQualityScore email reputation fraud score

=item IPQS_INVALID_EMAIL

    header   IPQS_INVALID_EMAIL eval:ipqs_invalid_email()
    score    IPQS_INVALID_EMAIL 1.0
    describe IPQS_INVALID_EMAIL whether the email appears valid

=item IPQS_NAME_UNKNOWN

    header   IPQS_NAME_UNKNOWN eval:ipqs_first_name()
    score    IPQS_NAME_UNKNOWN 0.1
    describe IPQS_NAME_UNKNOWN is the first name 'UNKNOWN'

=item IPQS_RECENT_ABUSE

    header   IPQS_RECENT_ABUSE eval:ipqs_recent_abuse()
    score    IPQS_RECENT_ABUSE 5.0
    describe IPQS_RECENT_ABUSE recently verified abuse

=item IPQS_FREQ_COMPLAIN

    header   IPQS_FREQ_COMPLAIN eval:ipqs_freq_complain()
    score    IPQS_FREQ_COMPLAIN 1.0
    describe IPQS_FREQ_COMPLAIN frequently unsubscribes

=item IPQS_VALID_BUT_SUSPECT

    header   __IPQS_SUSPECT         eval:ipqs_suspect()
    meta     IPQS_VALID_BUT_SUSPECT (!IPQS_INVALID_EMAIL && __IPQS_SUSPECT)
    score    IPQS_VALID_BUT_SUSPECT 0.5
    describe IPQS_VALID_BUT_SUSPECT valid result may be tainted

=item IPQS_DISPOSABLE

    header   IPQS_DISPOSABLE eval:ipqs_disposable()
    score    IPQS_DISPOSABLE 1.0
    describe IPQS_DISPOSABLE disposable or temporary mail service

=item IPQS_LEAKED

    header   IPQS_LEAKED eval:ipqs_disposable()
    score    IPQS_LEAKED 0.1
    describe IPQS_LEAKED has this address been leaked in a breach

=item IPQS_COMMON

    header   IPQS_COMMON eval:ipqs_common()
    score    IPQS_COMMON 0.0
    describe IPQS_COMMON address is from a common email provider

=back

=head2 HEADERS

Each of the data points returned by each call to our API is added in its own
header for each message checked by SpamAssassin. Please refer to the
L<IPQualityScore Email API
Documentation|https://www.ipqualityscore.com/documentation/email-validation/overview>
for detailed descriptions of these data points.


=head2 SUPPORT

For any questions, concerns, bug reports, or suggestions, please contact
IPQualityScore support: L<Support@IPQualityScore.com>

=head2 COPYRIGHT AND LICENSE

Copyright (C) 2023 IPQualityScore LLC

This is free software; you can redistribute it and/or modify it under
the terms of the Apache License 2.0. See the LICENSE file included
with this distribution for more information.

This plugin is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

=cut

#package Mail::SpamAssassin::Plugin::IPQualityScore;
package IPQualityScore;

our $VERSION = v1.0.0;

use strict;
use warnings;

use HTTP::Request;
use HTTP::Headers;
use JSON::MaybeXS;
use LWP::UserAgent;

use Mail::SpamAssassin::Plugin;
our @ISA = qw(Mail::SpamAssassin::Plugin);

sub dbg  { Mail::SpamAssassin::Logger::dbg("IPQualityScore: @_"); }
sub info { Mail::SpamAssassin::Logger::info("IPQualityScore: @_"); }

sub error {
    Mail::SpamAssassin::Logger::log_message( 'error', "IPQualityScore: @_" );
}

# plugin object constructor
sub new {
    my ( $class, $mailsa ) = @_;

    $class = ref($class) || $class;
    my $self = $class->SUPER::new($mailsa);
    bless( $self, $class );

    $self->register_eval_rule('ipqs_fraud_score');
    $self->register_eval_rule('ipqs_invalid_email');
    $self->register_eval_rule('ipqs_first_name');
    $self->register_eval_rule('ipqs_recent_abuse');
    $self->register_eval_rule('ipqs_freq_complain');
    $self->register_eval_rule('ipqs_suspect');
    $self->register_eval_rule('ipqs_disposable');
    $self->register_eval_rule('ipqs_leaked');
    $self->register_eval_rule('ipqs_common');

    $self->set_config( $mailsa->{conf} );

    return $self;
}

# parse config file
sub set_config {
    my ( $self, $conf ) = @_;
    my @cmds = ();

    # disable plugin
    push(
        @cmds,
        {
            setting => 'ipqs_disable',
            default => 0,
            type    => $Mail::SpamAssassin::Conf::CONF_TYPE_BOOL,
        }
    );

    # API key
    push(
        @cmds,
        {
            setting => 'api_key',
            code    => sub {
                my ( $self, $key, $value, $line ) = @_;
                unless ( defined $value && $value !~ /^$/ ) {
                    error('missing API key');
                    return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
                }

# NOTE: code will break if API key format changes from current 32-char alphanum string
                unless ( $value =~ /^[a-zA-Z0-9]{32}$/ ) {
                    error('invalid API key');
                    return $Mail::SpamAssassin::Conf::INVALID_VALUE;
                }
                $self->{api_key} = $value;
            }
        }
    );

    # ----
    # User config settings
    # ----

    # Fraud Score
    push(
        @cmds,
        {
            setting => 'fraud_score',
            default => 75.0,
            type    => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC
        }
    );

    # API request parameters

    # fast
    push(
        @cmds,
        {
            setting => 'fast',
            default => 0,
            type    => $Mail::SpamAssassin::Conf::CONF_TYPE_BOOL
        }
    );

    # timeout
    push(
        @cmds,
        {
            setting => 'timeout',
            default => 7,
            code    => sub {
                my ( $self, $key, $value, $line ) = @_;
                if ( defined $value ) {
                    unless ( $value >= 1 && $value <= 60 ) {
                        return $Mail::SpamAssassin::Conf::INVALID_VALUE;
                    }
                }
                else { $value = 7; }
                $self->{timeout} = $value;
            }
        }
    );

    # strictness
    push(
        @cmds,
        {
            setting => 'strictness',
            default => 0,
            code    => sub {
                my ( $self, $key, $value, $line ) = @_;
                if ( defined $value ) {
                    unless ( $value >= 0 && $value <= 2 ) {
                        return $Mail::SpamAssassin::Conf::INVALID_VALUE;
                    }
                }
                else { $value = 0; }
                $self->{strictness} = $value;
            }
        }
    );

    # abuse strictness
    push(
        @cmds,
        {
            setting => 'abuse_strictness',
            default => 0,
            code    => sub {
                my ( $self, $key, $value, $line ) = @_;
                if ( defined $value ) {
                    unless ( $value >= 0 && $value <= 2 ) {
                        return $Mail::SpamAssassin::Conf::INVALID_VALUE;
                    }
                }
                else { $value = 0; }
                $self->{abuse_strictness} = $value;
            }
        }
    );
    $conf->{parser}->register_commands( \@cmds );
}

# eval rules
# these are asynchronous rules; the logic is handled during check_dnsbl
sub ipqs_fraud_score {
    my ( $self, $pms ) = @_;
    dbg('ipqs_fraud_score eval rule');
    return undef;
}

sub ipqs_invalid_email {
    my ( $self, $pms ) = @_;
    dbg('ipqs_invalid_email eval rule');
    return undef;
}

sub ipqs_first_name {
    my ( $self, $pms ) = @_;
    dbg('ipqs_first_name eval rule');
    return undef;
}

sub ipqs_recent_abuse {
    my ( $self, $pms ) = @_;
    dbg('ipqs_recent_abuse eval rule');
    return undef;
}

sub ipqs_freq_complain {
    my ( $self, $pms ) = @_;
    dbg('ipqs_freq_complain eval rule');
    return undef;
}

sub ipqs_suspect {
    my ( $self, $pms ) = @_;
    dbg('ipqs_suspect eval rule');
    return undef;
}

sub ipqs_disposable {
    my ( $self, $pms ) = @_;
    dbg('ipqs_disposable eval rule');
    return undef;
}

sub ipqs_leaked {
    my ( $self, $pms ) = @_;
    dbg('ipqs_leaked eval rule');
    return undef;
}

sub ipqs_common {
    my ( $self, $pms ) = @_;
    dbg('ipqs_common eval rule');
    return undef;
}

# according to man 3 Mail::SpamAssassin::Plugin: "Called when DNSBL or other network lookups are being launched,
# implying current running priority of -100. This is the place to start your own asynchronously-started network
# lookups."
sub check_dnsbl {
    my ( $self, $opts ) = @_;

    my $pms  = $opts->{permsgstatus};
    my $conf = $pms->{conf};

    if ( $conf->{ipqs_disable} ) {
        info('plugin disabled.');
        return;
    }

    unless ( $conf->{api_key} ) {
        error(
'plugin will not work without a valid API key. Make sure to enter yours in ipqualityscore.cf.'
        );
        return;
    }

# Email can have multiple addresses in the From: header
# $status->get(header_name [, default_value]) should be called in list context since SA 4.0
    my @from     = $pms->get( 'From:addr', undef );
    my $num_addr = scalar @from;
    dbg("number of From: addresses assessed: $num_addr");

    # grab first address in From: list regardless
    my $target_address = $from[0];
    if ( $num_addr > 1 ) {

        # more than 1 "From:" address. Check the Sender: header...
        my @sender = $pms->get( 'Sender:addr', undef );
        if ( scalar @sender != 1 ) {
            error(
'message had multiple From: addresses but did not have exactly one Sender'
            );
            return;
        }

        # one Sender found, use this instead of "From:"
        $target_address = $sender[0];
    }

    # some simple validation of the target address
    # make sure there is at least something there
    unless ( defined $target_address && $target_address !~ /^$/ ) {
        error("failed to grab a From address");
        return;
    }

    # we have a valid target address, create the request
    my $header = HTTP::Headers->new(
        Content_Type => 'application/x-www-form-urlencoded',
        IPQS_KEY     => $conf->{api_key}
    );

    # API request optional parameters
    my $fast    = $conf->{fast};
    my $timeout = $conf->{timeout};
    my $s       = $conf->{strictness};
    my $as      = $conf->{abuse_strictness};
    dbg("fast: $fast");
    dbg("timeout: $timeout");
    dbg("strictness: $s");
    dbg("abuse strictness: $as");
    my $url = 'https://www.ipqualityscore.com/api/json/email';
    my $data =
"email=$target_address\&fast=$fast\&timeout=$timeout\&strictness=$s\&abuse_strictness=$as\&spamassassin=1";
    my $request  = HTTP::Request->new( 'POST', $url, $header, $data );
    my $ua       = LWP::UserAgent->new;
    my $response = $ua->request($request);

    # Process response
    if ( $response->is_success ) {
        my $result = decode_json( $response->decoded_content );
        unless ( $result->{success} ) {
            error( $result->{message} );
            foreach my $error ( $result->{errors}->@* ) {
                error($error);
            }
            return;
        }

        # Successful response
        my $request_id = $result->{request_id};

        # Assign values to the additional headers (those not attached to rules)
        $pms->set_tag( 'IPQSREQUESTID', $request_id );
        $pms->set_tag( 'IPQSTIMEDOUT',  $result->{timed_out} );
        $pms->set_tag( 'IPQSDELIV',     $result->{deliverability} );
        $pms->set_tag( 'IPQSCATCHALL',  $result->{catch_all} );
        $pms->set_tag( 'IPQSSMTPSCORE', $result->{smtp_score} );
        $pms->set_tag( 'IPQSOVERALL',   $result->{overall_score} );
        $pms->set_tag( 'IPQSGENERIC',   $result->{generic} );
        $pms->set_tag( 'IPQSDNSVALID',  $result->{dns_valid} );

        # -------------
        # Process rules
        # -------------

        # Fraud Score
        my $fraud_score = $result->{fraud_score};
        $pms->set_tag( 'IPQSFRAUDSCORE', "$fraud_score" );
        if ( $fraud_score >= $conf->{fraud_score} ) {
            $pms->test_log( "fraud_score: $fraud_score", 'IPQS_FRAUD_SCORE' );
            $pms->got_hit(
                'IPQS_FRAUD_SCORE',
                "$request_id ",
                ruletype => 'eval'
            );
        }

        # Valid Email
        if ( $result->{valid} ) {
            $pms->set_tag( 'IPQSVALIDEMAIL', 'TRUE' );
        }
        else {
            $pms->set_tag( 'IPQSVALIDEMAIL', 'FALSE' );
            $pms->test_log( 'invalid email', 'IPQS_INVALID_EMAIL' );
            $pms->got_hit(
                'IPQS_INVALID_EMAIL',
                "$request_id ",
                ruletype => 'eval'
            );
        }

        # First Name
        my $first_name = $result->{first_name};
        $pms->set_tag( 'IPQSFIRSTNAME', "$first_name" );
        if ( $result->{first_name} eq 'Unknown' ) {
            $pms->test_log( $result->{first_name}, 'IPQS_NAME_UNKNOWN' );
            $pms->got_hit(
                'IPQS_NAME_UNKNOWN',
                "$request_id ",
                ruletype => 'eval'
            );
        }
        else {
            dbg( $result->{first_name} );
        }

        # Recent Abuse
        if ( $result->{recent_abuse} ) {
            $pms->set_tag( 'IPQSRECENTABUSE', 'TRUE' );
            $pms->test_log( 'recent abuse verified', 'IPQS_RECENT_ABUSE' );
            $pms->got_hit(
                'IPQS_RECENT_ABUSE',
                "$request_id ",
                ruletype => 'eval'
            );
        }
        else {
            $pms->set_tag( 'IPQSRECENTABUSE', 'FALSE' );
        }

        # Frequent Complainer
        if ( $result->{frequent_complainer} ) {
            $pms->set_tag( 'IPQSFREQCOMPLAIN', 'TRUE' );
            $pms->test_log( "sender: $target_address", 'IPQS_FREQ_COMPLAIN' );
            $pms->got_hit(
                'IPQS_FREQ_COMPLAIN',
                "$request_id ",
                ruletype => 'eval'
            );
        }
        else {
            $pms->set_tag( 'IPQSFREQCOMPLAIN', 'FALSE' );
        }

        # Suspect
        if ( $result->{suspect} ) {
            $pms->set_tag( 'IPQSSUSPECT', 'TRUE' );
            $pms->test_log( 'mail server replying with temporary error',
                '__IPQS_SUSPECT' );
            $pms->got_hit(
                '__IPQS_SUSPECT',
                "$request_id ",
                ruletype => 'eval'
            );
        }
        else {
            $pms->set_tag( 'IPQSSUSPECT', 'FALSE' );
        }

        # Disposable
        if ( $result->{disposable} ) {
            $pms->set_tag( 'IPQSDISPOSABLE', 'TRUE' );
            $pms->test_log( 'disposable mail service', 'IPQS_DISPOSABLE' );
            $pms->got_hit(
                'IPQS_DISPOSABLE',
                "$request_id ",
                ruletype => 'eval'
            );
        }
        else {
            $pms->set_tag( 'IPQSDISPOSABLE', 'FALSE' );
        }

        # Leaked
        if ( $result->{leaked} ) {
            $pms->set_tag( 'IPQSLEAKED', 'TRUE' );
            $pms->test_log( 'email previously leaked', 'IPQS_LEAKED' );
            $pms->got_hit( 'IPQS_LEAKED', "$request_id ", ruletype => 'eval' );
        }
        else {
            $pms->set_tag( 'IPQSLEAKED', 'FALSE' );
        }

        # Common
        if ( $result->{common} ) {
            $pms->set_tag( 'IPQSCOMMON', 'TRUE' );
            $pms->test_log( 'common provider', 'IPQS_COMMON' );
            $pms->got_hit( 'IPQS_COMMON', "$request_id ", ruletype => 'eval' );
        }
        else {
            $pms->set_tag( 'IPQSCOMMON', 'FALSE' );
        }
    }

    # HTTP request was unsuccessful
    else {
        error( $response->status_line );
    }
}

1;
